/**CFile***********************************************************************

  FileName    [bcdCmd.c]

  PackageName [bcd]

  Synopsis    [Routines to detect bad cycles.]

  Author      [Aiguo Xie]

  Date        [10/3/1999]

  Copyright   [Copyright (c) 1999 Aiguo Xie. All rights reserved.]

******************************************************************************/

#include "fsmInt.h"
#include "bdd.h"
#include "ntm.h"
#include "img.h"
#include "sccInt.h"
#include "bcdInt.h"

int bcdHamming;
int bcdVerbose;
int bcdRecursionType;
int bcdMethod;
int bcdMachineFail;
int bcdBddCacheFlush;
int bcdBCgeneration;
int bcdFindAll;
int bcdRestrict;
int bcdNumCS;
char *bcdOutputFileName;
FILE *bcdOutputFile;

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
static int CommandComputeBadCycle(Hrc_Manager_t **hmgr, int argc, char **argv);
static int CommandPrintBadCycle(Hrc_Manager_t **hmgr, int argc, char **argv);

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis [Initializes the SCC package.]

  SideEffect []

******************************************************************************/
void Bcd_Init()
{
	Cmd_CommandAdd("compute_bc",CommandComputeBadCycle,0);
	Cmd_CommandAdd("print_bc",CommandPrintBadCycle,0);
}

/**Function********************************************************************

  Synopsis [Ends the BCD package.]

  SideEffect []

******************************************************************************/
void Bcd_End()
{
}

/**Function********************************************************************

  Synopsis [Implements the compute_bc command.]

  Description []
  SideEffect []
  AlsoSee [CommandPrintBadCycle]

******************************************************************************/
static int CommandComputeBadCycle(Hrc_Manager_t **hmgr, int argc, char **argv)
{
	Fsm_Fsm_t *fsm = Fsm_HrcManagerReadCurrentFsm(*hmgr);
	long initialTime, finalTime;
	char c;
	int i;

	sccMddList *bcdCS = NULL; /* for test only */

	bcdHamming = 0;
	bcdVerbose = 0;
	bcdRecursionType = BCD_BACKWARD_ONLY;
	bcdMethod = BCD_XB_METHOD;
	bcdMachineFail = BCD_FALSE; /* for test only */
	bcdBddCacheFlush = BCD_FALSE;
	bcdBCgeneration = BCD_RANDOM;
	bcdFindAll = BCD_FALSE;
	bcdOutputFile = NULL;
	bcdOutputFileName = NULL;
	bcdRestrict = BCD_FALSE;
	util_getopt_reset();

	while((c=util_getopt(argc, argv, "v:m:o:h:cdabfr")) != EOF) {
		switch (c) {
			case 'o':
				bcdOutputFileName = 
				(char *)calloc(1,sizeof(strlen(util_optarg)+10));
				*bcdOutputFileName = '\0'; 
				sprintf(bcdOutputFileName, "exp_res/%s",util_optarg);
				bcdOutputFile = fopen(bcdOutputFileName, "w");
				if (bcdOutputFile == NULL) {
					fprintf(vis_stderr, 
						"cannot open %s.\n",
						bcdOutputFileName);
					return 1;
				}
				break;
			case 'h':
				bcdHamming = atoi(util_optarg);
				break;
			case 'v':
				bcdVerbose = atoi(util_optarg);
				break;
			case 'm':
				if (!strcmp(util_optarg, "el")) 
					bcdMethod = BCD_EL_METHOD;
				else if (!strcmp(util_optarg, "xb"))
					bcdMethod = BCD_XB_METHOD;
				else {
					fprintf(vis_stderr,
					"** bcd error: undefined detection 
					method.\n");
					goto usage;
				}
				break;
			case 'b':
				bcdRecursionType = BCD_BACKWARD_ONLY;
				break;
			case 'f':
				bcdRecursionType = BCD_BACKWARD_FORWARD;
				break;
			case 'c':
				bcdBddCacheFlush = BCD_TRUE;
				break;
			case 'd':
				bcdMachineFail = BCD_TRUE;
				break;
			case 'a':
				bcdFindAll = BCD_TRUE;
				break;
			case 'r':
				bcdRestrict = BCD_TRUE;
				break;
			default:
				goto usage;
		}
	}
	/* alpha version: no command-line options allowed */ 

	if (bcdOutputFile) {
		fprintf(bcdOutputFile, "command:");
		fprintf(vis_stderr, "\ncommand:");
		for (i = 0; i<argc; i++) {
			fprintf(bcdOutputFile, " %s", argv[i]);
			fprintf(vis_stderr, " %s", argv[i]);
		}
		fprintf(bcdOutputFile, "\n");
		fprintf(vis_stderr, "\n");
	}
	/* read in cycle sets, see e.g. [1] of README */
	if (!bcdCS) {
		bcdCS = bcdReadCycleSets(fsm);
		if (!bcdCS) {
			fprintf(vis_stderr, "** bcd error: failure to read in cycle states.\n");
			return 1;
		}
	}
	if (bcdMethod == BCD_XB_METHOD) {
		if (!(fsm->reachabilityInfo.reachableStates)) {
			fprintf(vis_stderr, "** scc warning: 
				reachability analysis not performed yet.\n");
			fprintf(vis_stderr, "** scc warning: 
				it's being performed (IWLS95 method) ...\n");
			if (Cmd_CommandExecute(hmgr, "compute_reach"))
				return 1;
			fprintf(vis_stderr, "done.\n");
		}
		initialTime = util_cpu_time();
		if (bcd_XB(fsm, fsm->reachabilityInfo.reachableStates, bcdCS))
			return 1;
		finalTime = util_cpu_time();
	} else if (bcdMethod == BCD_EL_METHOD) {
		initialTime = util_cpu_time();
		if (bcd_EL(fsm, fsm->reachabilityInfo.reachableStates, bcdCS))
			return 1;
		finalTime = util_cpu_time();
	} else {
		fprintf(vis_stderr, "** bcd error: unrecoginized detection method).\n");
		return 1;
	}
	bcd_PrintResult(finalTime-initialTime, bcdMethod);
	if (bcdOutputFile) 
		fclose(bcdOutputFile);
	if (bcdOutputFileName) 
		free(bcdOutputFileName);
	return 0;

	usage:
	fprintf(vis_stderr,"compute_bc [-v #] [-b] [-f] [-d] [-c]\n");	
	fprintf(vis_stderr,"	-v # \t\t verbosity level\n");
	fprintf(vis_stderr,"    -m el|xb \t\t select a detection method\n");
	fprintf(vis_stderr,"    -o file \t\t specify an output file\n");
	fprintf(vis_stderr,"	-c \t\t clean up cache bdds\n");
	fprintf(vis_stderr,"	-d \t\t make up a failing version\n");
	fprintf(vis_stderr,"	-b \t\t by backward recusion (xb only)\n");
	fprintf(vis_stderr,"	-f \t\t by back- and forward- recusion (xb only)\n");
	fprintf(vis_stderr,"	-a \t\t find all bad sccs (xb only)\n");
	return 1;
}

/**Function********************************************************************

  Synopsis [Implement the print_scc command.]

  Description []
  SideEffect []
  AlsoSee [CommandComputeScc]

******************************************************************************/
static CommandPrintBadCycle(Hrc_Manager_t **hmgr,int argc,char **argv)
{
	fprintf(vis_stderr, 
		"** bcd error: command not completely implemented yet.");
	return 0;
}

void bcd_PrintResult(elapseTime, method)
long elapseTime;
int method;
{
	char *method_name;
	if (method == BCD_EL_METHOD) {
		method_name = 
		(char *)calloc(1,sizeof(char)*(strlen(BCD_EL_METHOD_NAME)+1));
		method_name[0] = '\0';
		sprintf(method_name, BCD_EL_METHOD_NAME);
	} else if (method == BCD_XB_METHOD) {
		method_name =
		(char *)calloc(1,sizeof(char)*(strlen(BCD_XB_METHOD_NAME)+1));
		sprintf(method_name, BCD_XB_METHOD_NAME);
	} else {
		fprintf(vis_stderr, "** bcd error: internal -- unknown detection method.\n");
	}
	fprintf(vis_stderr, "\n %-20s%s\n", "analysis method:", method_name); 
	fprintf(vis_stderr, "\n %-20s%g\n", "analysis time:",
				(double)elapseTime/1000.0);
	if (bcdOutputFile) {
		fprintf(bcdOutputFile, "\n %-20s%s\n", "analysis method:", method_name); 
		fprintf(bcdOutputFile, " %-20s%g\n", "analysis time:",
				(double)elapseTime/1000.0);
	}
	fprintf(vis_stderr, "========== END OF BCD ==========\n\n");
}

char *bcd_strSave(char *s) {
	char *t;
	t = (char *)calloc(1,sizeof(char)*strlen(s)+1);
	*t = '\0';
	sprintf(t, "%s", s);
	return t;
}
