
/**CFile**************************************************************************
  FileName    [bcdXB.c]
  PackageName [bcd]
  Synopsis    [Routines to detect bad cycles based on Xie and Beerel's SCC method]
  Author      [Aiguo Xie]
  Date        [10/3/1999]
  Copyright   [Copyright (c) 1999 Aiguo Xie. All rights reserved]
*********************************************************************************/

#include "fsmInt.h"
#include "bdd.h"
#include "ntm.h"
#include "img.h"
#include "imgInt.h"
#include "sccInt.h"
#include "bcdInt.h"

int bcdRecurDepth, maxBcdRecurDepth;

/* the algorithm:
 *
 * Let {c1, c2, ..., cn} be the set of characteristic functions of
 * the set of cycle sets. Let {f1, f2, ..., fn} be the set of
 * the characteristic functions of the set of complements of the
 * cycle sets. A bad cycle b must touches all the complements of
 * the cycle stes. That is, b^fi != 0 for all i = 1, ..., n.
 *
 * Let h be an scc that intersects at least one state with f1. Then,
 * h is bad if and only if h^fi != 0 for all i = 2, ..., n. Let H
 * = {h1, h2, ..., hm} be the set of sccs interesecting f1. The
 * detection problem passes if and only if none of sccs in H is
 * bad.
 */

static foundSCCid;

static void bcd_XB_Init(Fsm_Fsm_t *fsm);

int bcd_XB(Fsm_Fsm_t *fsm, mdd_t *reachable, sccMddList *csList)
{
    sccMddList *mddListCurs, *csCmplList, *lastListElement, *newListElement, *nextListElement;
    sccMddList *remainingCmplList;
    mdd_t *f1, *x, *s, *Bs, *base;
    mdd_t *tempMdd, *remaining, *S;
    int detect_flag = BCD_FAIL;

    bcd_XB_Init(fsm);

    foundSCCid =0;
    if (!csList) {
        fprintf(vis_stderr, "machine fails <= no cycle sets specified.\n");
        return 0;
        /* assuming there is at least one scc in the reachability
         * graph
         */
    }

    /* complementing the cycle sets */
    mddListCurs = csList;
    lastListElement = NULL;
    csCmplList = NULL;
    while (mddListCurs) {
        newListElement = (sccMddList *)calloc(1,sizeof(sccMddList));
        tempMdd = mdd_not(mddListCurs->val);
        newListElement->val = mdd_and(tempMdd, reachable,1,1);
        mdd_free(tempMdd);
	newListElement->id = mddListCurs->id;
if(bcdVerbose) {
	newListElement->size = bcd_numStates(fsm,newListElement->val);
}
        if (csCmplList == NULL) {
            csCmplList = newListElement;
        }
        if (lastListElement)
            lastListElement->next = newListElement;
        lastListElement = newListElement;
        mddListCurs = mddListCurs->next;
    }

    /* take out the mdd of the first complement */
    f1 = csCmplList->val;

if(bcdVerbose){
    fprintf(vis_stderr, "take out the complement (%g states) of the %d-th cycle set to compute sccs.\n", bcd_numStates(fsm, f1), csCmplList->id);
}

    remainingCmplList = csCmplList->next;

    /* iteratively compute all scc's passing through f1 */
    remaining = mdd_dup(f1);
    S = mdd_dup(reachable);
    while (!(mdd_is_tautology(remaining,0) || mdd_is_tautology(S, 0))) {

        if (bcdRecurDepth > 0) {
            fprintf(vis_stderr,
                " ** bcd_error: inconsistent recursion depth.\n");
            return 1;
        }

        s = scc_TakeAState(fsm, remaining);
        Bs = scc_BackwardReachableStates(fsm,s,S);
        base = mdd_and(remaining,Bs,1,1);

if (bcdVerbose>0) {
	fprintf(vis_stderr, "take a state s from the complement.\n");
	fprintf(vis_stderr, "size(Bs) = %g\n", bcd_numStates(fsm, Bs));
	fprintf(vis_stderr, "size(base) = %g\n", bcd_numStates(fsm, base));
}

        if (sccRecursionType == BCD_BACKWARD_ONLY) {
	    if (bcdVerbose>0)
	    	fprintf(vis_stderr, "\t recursive backward bcd...\n");
            if (bcd_XB_recur_backwardOnly(fsm,s,Bs,base,csCmplList, &detect_flag))
                return 1;
            if (detect_flag == BCD_SUCCEED) {
                /* find a bad cycle */
		if (bcdFindAll == BCD_FALSE) {
			mdd_free(s);
			mdd_free(Bs);
			mdd_free(base);
			break;
		}
            }
        } else if (sccRecursionType == BCD_BACKWARD_FORWARD) {
            if (bcd_XB_recur_backwardForward(fsm, s, Bs, base, csCmplList))
                return 1;
            if (detect_flag == BCD_SUCCEED) {
                /* find a bad cycle */
		if (bcdFindAll == BCD_FALSE) {
			mdd_free(s);
			mdd_free(Bs);
			mdd_free(base);
			break;
	    }
            }
        } else {
            fprintf(vis_stderr, "unknown recursion type.\n");
            return 1;
        }
        x = scc_MddOrDestroy(s, Bs, 1, 1, 1, 1);
        remaining = scc_MddAndDestroy(remaining, x, 1,0,1,0);
        S = scc_MddAndDestroy(S, x, 1, 0, 1, 1);
    }
    if (detect_flag == BCD_SUCCEED)
        fprintf(vis_stderr, "vis has detected a bad cycle.\n");
    else
    	fprintf(vis_stderr, "vis detects no bad cycle.\n");

    /* cleanup before leave */
    mdd_free(S);
    mdd_free(remaining);
    lastListElement = csCmplList;
    while (lastListElement) {
	nextListElement = lastListElement->next;
	mdd_free(lastListElement->val);
	free(lastListElement);
	lastListElement = nextListElement;
    }

    return 0;
}


/**Function********************************************************************

  Synopsis [Recursively detects bad cycles of the backward set of a given state.]

  Description [Recusively decompose the backward set of a state seed into
  SCCs and non-SCC states. Each call to the procedure removes the SCC
  containing the state seed (seed may not be in any SCC at the top call
  of the procedure). Next, all the states that can reach the removed SCC
  or the seed without passing any states of some undiscovered SCCs are
  iteratively identified and removed. Then, the states of the remaining
  set that can reach any of the states just removed are identified.
  This set of states is call the backward frontier set. While the
  backward frontier set is not empty, the procedure picks a state
  out of it and compute its backward set, and then calls itself.
  When the call finishes, the backward frontier set gets upated and
  the process repeats until the frontier set is empty.
  ]

  SeeAlso []

******************************************************************************/
int
bcd_XB_recur_backwardOnly (Fsm_Fsm_t *fsm, mdd_t *seed, mdd_t *stateSpace,
    mdd_t *base, sccMddList *csCmplList, int *detect_flag)
{
    mdd_t *Bs, *Fs, *xx, *x, *y, *z, *s, *R, *Bf;
    mdd_t *my_base, *this_base, *tempMdd;

    /*
    *detect_flag = BCD_FAIL;
    */
    bcdRecurDepth++;
    if (bcdRecurDepth > maxBcdRecurDepth)
        maxBcdRecurDepth = bcdRecurDepth;

if (bcdVerbose>0)
    fprintf(vis_stderr, "\trecur depth: %d\n", bcdRecurDepth);

    s = mdd_dup(seed); 
    Bs = mdd_dup(stateSpace);
    my_base = mdd_dup(base);
    Fs = scc_ForwardReachableStates(fsm, s, Bs);

if (bcdVerbose>0)
    fprintf(vis_stderr, "\tsize(Fs) = %g\n", bcd_numStates(fsm, Fs));

    /* take the intersection */
    x = scc_MddAndDestroy(Fs,Bs,1,1,1,0);

    /* check if the intersection is empty */
    if (!mdd_is_tautology(x,0)) {

        xx = scc_FwdEscapingImage(fsm,x);
        if (mdd_is_tautology(xx,0))
            scc_ReportSCC(fsm,x,TERMINAL, BACKWARD, 0);
        else
            scc_ReportSCC(fsm,x,INTERNAL, BACKWARD, 0);
        mdd_free(xx);

	tempMdd = mdd_and(x, base, 1,1);
	if (!mdd_is_tautology(tempMdd,0)) {
		fprintf(vis_stderr, "found an scc (#%4d)...checking...", ++foundSCCid);
        	if (bcd_allIntersect(fsm, x, csCmplList)) {
			fprintf(vis_stderr, "bad\n");
        	    	*detect_flag = BCD_SUCCEED; /* found a bad cycle */
			if (bcdFindAll == BCD_FALSE) {
	    			bcdRecurDepth--;
            			return 0;
			}
        	} else
			fprintf(vis_stderr, "good\n");
	} else {
		fprintf(vis_stderr, "found an scc...not passing any state of base...good\n");
if (bcdVerbose>0) {
		fprintf(vis_stderr, "\tfound an scc (%g states)\n",
		bcd_numStates(fsm, x));
		fprintf(vis_stderr, "\tnot covering any state of the base (part of the complement): ignored.\n");
}
	}
	free(tempMdd);
        my_base = scc_MddAndDestroy(my_base, x, 1, 0, 1, 0);
        R = scc_MddAndDestroy(Bs,x,1,0,1,0);
    } else {
        scc_ReportSCC(fsm,s,NONSCC, BACKWARD, 0);
        my_base = scc_MddAndDestroy(my_base, s, 1, 0, 1, 0);
        R = scc_MddAndDestroy(Bs,s,1,0,1,0);
    }
    /*
     * iteratively identify all non-SCCs reaching the states
     * just removed
     */
    y = scc_MddOrDestroy(s,x,1,1,1,1);
    z = scc_NonSCCiteration(fsm,y,R);
    if (!mdd_is_tautology(z,0)) {
        scc_ReportSCC(fsm, z, NONSCC, BACKWARD, 0);
        my_base = scc_MddAndDestroy(my_base, z, 1, 0, 1, 0);
        R = scc_MddAndDestroy(R,z,1,0,1,0);
    }
    /* compute the backward frontier set */
    xx = scc_MddOrDestroy(y,z,1,1,1,1);
    Bf = scc_RestrictedBwdImage(fsm,xx,R);
    mdd_free(xx);

    /* emptiess of Bf <==> emptiness of R */
    while (!mdd_is_tautology(R,0) && !mdd_is_tautology(my_base,0) ) {
        s  = scc_TakeAState(fsm, Bf);
        Bs = scc_BackwardReachableStates(fsm,s,R);
        this_base = scc_MddAndDestroy(my_base, Bs, 1, 1, 0, 0);
        if (bcd_XB_recur_backwardOnly(fsm,s,Bs,this_base,csCmplList,detect_flag)) {
	    bcdRecurDepth--;
            return 1;
        }
	mdd_free(this_base);
	y = scc_MddOrDestroy(s,Bs,1,1,1,1);
        my_base = scc_MddAndDestroy(my_base, y, 1, 0, 1, 0);
        R  = scc_MddAndDestroy(R, y,1,0,1,0);
        Bf = scc_MddAndDestroy(Bf,y,1,0,1,1);
        if (*detect_flag == BCD_SUCCEED && bcdFindAll == BCD_FALSE)
            break;
    }

    /* if Bf is not empty, something is wrong */
    if (!mdd_is_tautology(Bf,0) && 
	!mdd_is_tautology(my_base, 0) && *detect_flag != BCD_SUCCEED) {
        fprintf(vis_stderr,
            "** bcd_error: something is wrong with the code.\n");
        return 1;
    }
    mdd_free(Bf);
    mdd_free(R);
    bcdRecurDepth--;
    return 0;
}

static void bcd_XB_Init(Fsm_Fsm_t *fsm)
{
    mdd_manager *mddManager;

    fsm->imageInfo = Fsm_FsmReadOrCreateImageInfo(fsm,1,1);

    bcdRecurDepth = 0;
    maxBcdRecurDepth = bcdRecurDepth;

    if (bcdBddCacheFlush == BCD_TRUE) {
    	mddManager = 
    	Part_PartitionReadMddManager(fsm->imageInfo->functionData.mddNetwork);
	fprintf(vis_stderr, "flushing bdd cache...");
    	cuddCacheFlush(mddManager);
	fprintf(vis_stderr, "done\n"); 
    }
}

int bcd_XB_recur_backwardForward(Fsm_Fsm_t *fsm, mdd_t *s, mdd_t *Bs, mdd_t *base, mdd_t *csCmplList)
{
	fprintf(vis_stderr, "bcd_XB_recur_backForward not defined yet.\n");
	return 1;
}

int bcd_allIntersect(Fsm_Fsm_t *fsm, mdd_t *B, sccMddList *list) {
	sccMddList *curs;
	mdd_t *tempMdd;

	double num;
	mdd_manager *mddManager;

	mddManager = 
	Part_PartitionReadMddManager(fsm->imageInfo->functionData.mddNetwork);

if (bcdVerbose>0) {
	num = mdd_count_onset(mddManager, B, fsm->fsmData.presentStateVars);
	fprintf(vis_stderr, 
	"checking intersection of an scc (%g states) with cycle sets...\n",num);
}
	curs = list;
	while (curs) {
		/*
		tempMdd = mdd_and(B, curs->val, 1, 1);
		*/
		tempMdd = bdd_intersects_binary(B, curs->val);
		if (mdd_is_tautology(tempMdd, 0)) {
			mdd_free(tempMdd);
			if (bcdVerbose>0)
				fprintf(vis_stderr, "does not intersect with a complement of a cycle set.\n");
			return 0;
		}
		mdd_free(tempMdd);
		curs = curs->next;
	}
	if (bcdVerbose>0)
		fprintf(vis_stderr, "intersects with all complements of cycle sets.\n");
	return 1;
}

double bcd_numStates(Fsm_Fsm_t *fsm, mdd_t *x) {
	mdd_manager *mddManager =
	Part_PartitionReadMddManager(fsm->imageInfo->functionData.mddNetwork);
	return mdd_count_onset(mddManager, x, fsm->fsmData.presentStateVars);
}
