/**CFile***********************************************************************

  FileName    [sccCmd.c]

  PackageName [scc]

  Synopsis    [Routines to decompose state space into SCCs.]

  Author      [Aiguo Xie]

  Copyright   [Copyright (c) 1999 The University of Southern California.
  All rights reserved.]

******************************************************************************/

#include "fsmInt.h"
#include "bdd.h"
#include "ntm.h"
#include "img.h"
#include "sccInt.h"

int sccVerbose;
int sccRecursionType;
int sccMethod;
sccMddList *sccInternalSCCList, *sccExternalSCCList;

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
static int CommandComputeScc(Hrc_Manager_t **hmgr, int argc, char **argv);
static int CommandPrintScc(Hrc_Manager_t **hmgr, int argc, char **argv);

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis [Initializes the SCC package.]

  SideEffect []

******************************************************************************/
void Scc_Init()
{
	Cmd_CommandAdd("compute_scc",CommandComputeScc,0);
	Cmd_CommandAdd("print_scc",CommandPrintScc,0);
}

/**Function********************************************************************

  Synopsis [Ends the SCC package.]

  SideEffect []

******************************************************************************/
void Scc_End()
{
}

/**Function********************************************************************

  Synopsis [Implements the compute_scc command.]

  Description []
  SideEffect []
  AlsoSee [CommandPrintScc]

******************************************************************************/
static int CommandComputeScc(Hrc_Manager_t **hmgr, int argc, char **argv)
{
	Fsm_Fsm_t *fsm = Fsm_HrcManagerReadCurrentFsm(*hmgr);
	long initialTime, finalTime;
	char c;
	int recordSCC;

	recordSCC = 0;
	sccVerbose = 0;
	sccRecursionType = BACKWARD_ONLY;
	sccMethod = USC_SCC_METHOD;
	sccInternalSCCList = NULL;
	sccExternalSCCList = NULL;
	util_getopt_reset();
	while((c=util_getopt(argc, argv, "v:m:bf:r")) != EOF) {
		switch (c) {
			case 'v':
				sccVerbose = atoi(util_optarg);
				break;
			case 'b':
				sccRecursionType = BACKWARD_ONLY;
				break;
			case 'f':
				sccRecursionType = BACKWARD_FORWARD;
				break;
			case 'm':
				if (!strcmp(util_optarg, "ucb")) 
					sccMethod = UCB_SCC_METHOD;
				else if (!strcmp(util_optarg, "usc"))
					sccMethod = USC_SCC_METHOD;
				else {
					fprintf(vis_stderr,
					"** scc error: undefined decompose 
					method.\n");
					goto usage;
				}
				break;
			case 'r':
				fprintf(vis_stderr, "scc-recorde flag is set.\n");
				recordSCC = 1;
				break;
			default:
				goto usage;
		}
	}
	/* alpha version: no command-line options allowed */ 
        fprintf(vis_stderr, "computing all sccs...");
	if (sccMethod == USC_SCC_METHOD) {
		if (!(fsm->reachabilityInfo.reachableStates)) {
			/* reachability not done yet */
			fprintf(vis_stderr, "** scc warning: 
				reachability analysis not performed yet.\n");
			fprintf(vis_stderr, "** scc warning: 
				it's being performed (IWLS95 method) ...\n");
			if (Cmd_CommandExecute(hmgr, "compute_reach"))
				return 1;
			fprintf(vis_stderr, "done.\n");
		}
		initialTime = util_cpu_time();
		if (scc_Decompose(fsm, fsm->reachabilityInfo.reachableStates,recordSCC))
			return 1;
		finalTime = util_cpu_time();
	} else if (sccMethod == UCB_SCC_METHOD) {
		/*
		initialTime = util_cpu_time();
		*/
		if (scc_Decompose_TC(fsm, &initialTime, &finalTime))
			return 1;
		/*
		finalTime = util_cpu_time();
		*/
	} else {
		fprintf(vis_stderr, "scc error: internal problem (decompose method unrecoginized).\n");
		return 1;
	}
	fprintf(vis_stderr, "done.\n");
	scc_PrintResult(finalTime-initialTime, sccMethod);

	return 0;

	usage:
	fprintf(vis_stderr,"compute_scc [-v #] [-b] [-f]\n");	
	fprintf(vis_stderr,"	-v # \t\t verbosity level\n");
	fprintf(vis_stderr,"    -m usc|ucb \t\t select a decompose method.\n");
	fprintf(vis_stderr,"	-b \t\t by backward recusion (usc only)\n");
	fprintf(vis_stderr,"	-f \t\t by back- and forward- recusion (usc only)\n");
	return 1;
}

/**Function********************************************************************

  Synopsis [Implement the print_scc command.]

  Description []
  SideEffect []
  AlsoSee [CommandComputeScc]

******************************************************************************/
static CommandPrintScc(Hrc_Manager_t **hmgr,int argc,char **argv)
{
	fprintf(vis_stderr, 
		"** scc error: command not completely implemented yet.");
	return 0;
}
