/**CFile***********************************************************************

  FileName    [sccReach.c]

  PackageName [scc]

  Synopsis    [Routines to perform reachability on the FSM structure.]

  Author      [Aiguo Xie]

  Copyright   [Copyright (c) 1999 The University of Southern California.
  All rights reserved.]

******************************************************************************/

#include "fsmInt.h"
#include "bdd.h"
#include "ntm.h"
#include "img.h"
#include "imgInt.h"
#include "sccInt.h"

#define FSM_MDD_DONT_FREE 0  /* flag to indicate that bdd should not be freed. */
#define FSM_MDD_FREE 1       /* flag to indicate that bdd should be freed.*/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis [Computes the set of the forward reachables of an FSM.]

  Description [Computes the set of forward reachable states from a 
  designated set of initial states which may not match that of the 
  FSM itself] 

******************************************************************************/
mdd_t *
scc_ForwardReachableStates(Fsm_Fsm_t *fsm, mdd_t *initialStates, mdd_t *bound)
{
  /* BFS variables */
  Img_ImageInfo_t *imageInfo = NIL(Img_ImageInfo_t);
  mdd_t           *reachableStates;/* reachable states */
  mdd_t           *fromLowerBound; /* new states in each iteration */
  mdd_t           *fromUpperBound; /* set to reachable states */
  mdd_t           *image;          /* image computed at each iteration */
  mdd_t           *toCareSet;      /* the complement of the reached set */

  mdd_manager     *mddManager = Fsm_FsmReadMddManager(fsm); /* mdd manager */
  Ntk_Network_t   *network = fsm->network; /* network */
  bdd_reorder_type_t currentMethod =
    Ntk_NetworkReadDynamicVarOrderingMethod(network); /* current reordering
						       * method */
  int firstReorderFlag = FALSE;
                         /* a flag to indicate the start of the computation */
  int reachableStateSetSize; /* bdd size of the reached set */
  array_t *mintermVarArray; /* array of present state variables */

  int nvars;
  FsmHdMintSizeStats_t *stats;
  int depth=0;
  
  /* initializations */
  /* if fsm is a submachine, takes realPsVars to count minterm correctly. */
  mintermVarArray = (Fsm_FsmReadRealPresentStateVars(fsm) != NIL(array_t)) ? Fsm_FsmReadRealPresentStateVars(fsm) :  Fsm_FsmReadPresentStateVars(fsm);

  /* compute number of present state variables */
  nvars = ComputeNumberOfBinaryStateVariables(mddManager, mintermVarArray);
  assert(nvars > 0);

  /*
  if (!(fsm->imageInfo == NIL(Img_ImageInfo_t))) {
  	free(fsm->imageInfo);
  	fsm->imageInfo = NIL(Img_ImageInfo_t);
  }
  imageInfo = Fsm_FsmReadOrCreateImageInfo(fsm,0,1);
  */
  imageInfo = fsm->imageInfo;

  /* initializations */
  reachableStateSetSize = 0;
  stats = NULL;

  if (initialStates == NIL(mdd_t)) {
    fprintf(vis_stdout,
      "** rch error: No initial states specified. 
      No reachability will be performed.\n");
    return NIL(mdd_t);
  }

  /* start reached set with the empty set */
  reachableStates = scc_mdd_zero(fsm);

  fromLowerBound = mdd_dup(initialStates);
  fromUpperBound = initialStates;
  
  /* Main loop of reachability computation. */
  /* Iterate until fromLowerBound is empty or all states are reached. */
  while ((!mdd_is_tautology(fromLowerBound, 0)) &&
	 (!mdd_is_tautology(reachableStates, 1))) {
    
    /* if it is the print step, print out the iteration and the */
    #ifdef PRINT
    if (printStep && ((depth % printStep) == 0)) {
      int step;
      /* for BFS, preserve earlier format of printing */
      step = Fsm_FsmGetReachableDepth(fsm)+depth;
      PrintStatsPerIteration(approxFlag,
			     nvars,
			     step,
			     stats,
			     mddManager,
			     reachableStates,
			     fromLowerBound,
			     mintermVarArray);
    }
    #endif

    #ifdef REORDER
    if (reorderFlag && !firstReorderFlag &&
	(reachableStateSetSize >= reorderThreshold )){
      firstReorderFlag = TRUE;
      (void) fprintf(vis_stdout, "Dynamic variable ordering forced ");
      (void) fprintf(vis_stdout, "with method sift\n");
      Ntk_NetworkSetDynamicVarOrderingMethod(network,
		     BDD_REORDER_SIFT, BDD_REORDER_VERBOSITY_DEFAULT);
      bdd_reorder(Ntk_NetworkReadMddManager(network));
      Ntk_NetworkSetDynamicVarOrderingMethod(network, currentMethod,
				     BDD_REORDER_VERBOSITY_DEFAULT);
    }
    #endif

    /* careSet for image computation is the set of all states not reached
     * so far.
     */
    toCareSet = mdd_not(reachableStates);

    /*
     * Image of some set between lower and upper, where we care
     * about the image only on unreached states.
     */

    /* compute the image of this iteration */
    image = Img_ImageInfoComputeFwdWithDomainVars(imageInfo,
						    fromLowerBound,
						    fromUpperBound,
						    toCareSet);
    /* free the used bdds */
    mdd_free(toCareSet);
    
    /* New lower bound is the states just reached. */
    mdd_free(fromLowerBound);

    image = scc_MddAndDestroy(image,bound,1,1,1,0);
    fromLowerBound = mdd_and(image, reachableStates, 1, 0);
    depth++;

    mdd_free(image);
    image = NULL;
    
    /* New set of reachable states is old set plus new states in image. */
    reachableStates = (mdd_t *)AddStates(fromLowerBound, reachableStates,
			  FSM_MDD_DONT_FREE, FSM_MDD_FREE);
    fromUpperBound = reachableStates;

    if (mdd_is_tautology(fromLowerBound, 0)) depth--;
      
  } /* end of while loop (while frontier != 0) */

  #ifdef PRINT
  if (mdd_is_tautology(fromLowerBound, 0) || 
      mdd_is_tautology(reachableStates, 1)) {

    if (mdd_is_tautology(reachableStates, 1) && 
       !mdd_is_tautology(fromLowerBound, 0) &&
	printStep && ((depth % printStep) == 0) ) {
      int step;
      /* for BFS, preserve earlier format of printing */
      step = 
      (shellFlag) ? array_n(onionRings) : Fsm_FsmGetReachableDepth(fsm)+depth;
      PrintStatsPerIteration(approxFlag,
			     nvars,
			     step,
			     stats,
			     mddManager,
			     reachableStates,
			     fromLowerBound,
			     mintermVarArray);
    }
  }
  #endif
  
  mdd_free(fromLowerBound);
  /*
  mdd_free(initialStates);
  */

  return reachableStates;
} /* end of scc_computeReachableStates() */

/**Function********************************************************************

  Synopsis [Computes the set of the backward reachables of an FSM.]

  Description [Computes the set of backward reachable states from a 
  designated set of initial states which may not match that of the 
  FSM itself] 

******************************************************************************/
mdd_t *
scc_BackwardReachableStates(Fsm_Fsm_t *fsm, mdd_t *initialStates, mdd_t *bound)
{
  /* BFS variables */
  Img_ImageInfo_t *imageInfo = NIL(Img_ImageInfo_t);
  mdd_t           *reachableStates;/* reachable states */
  mdd_t           *fromLowerBound; /* new states in each iteration */
  mdd_t           *fromUpperBound; /* set to reachable states */
  mdd_t           *image;          /* image computed at each iteration */
  mdd_t           *toCareSet;      /* the complement of the reached set */

  mdd_manager     *mddManager = Fsm_FsmReadMddManager(fsm); /* mdd manager */
  Ntk_Network_t   *network = fsm->network; /* network */
  bdd_reorder_type_t currentMethod =
    Ntk_NetworkReadDynamicVarOrderingMethod(network); /* current reordering
						       * method */
  int firstReorderFlag = FALSE;
                         /* a flag to indicate the start of the computation */
  int reachableStateSetSize; /* bdd size of the reached set */
  array_t *mintermVarArray; /* array of present state variables */
  
  int nvars;
  FsmHdMintSizeStats_t *stats;
  int depth=0;

  /* initializations */
  /* if fsm is a submachine, takes realPsVars to count minterm correctly. */
  mintermVarArray = (Fsm_FsmReadRealPresentStateVars(fsm) != NIL(array_t)) ? Fsm_FsmReadRealPresentStateVars(fsm) :  Fsm_FsmReadPresentStateVars(fsm);

  /* compute number of present state variables */
  nvars = ComputeNumberOfBinaryStateVariables(mddManager, mintermVarArray);
  assert(nvars > 0);

  /*
  if (!(fsm->imageInfo == NIL(Img_ImageInfo_t))) {
  	free(fsm->imageInfo);
  	fsm->imageInfo = NIL(Img_ImageInfo_t);
  }
  imageInfo = Fsm_FsmReadOrCreateImageInfo(fsm,1,0);
  */
  imageInfo = fsm->imageInfo;

  /* initializations */
  reachableStateSetSize = 0;
  stats = NULL;

  if (initialStates == NIL(mdd_t)) {
    fprintf(vis_stdout,
      "** rch error: No initial states specified. 
      No reachability will be performed.\n");
    return NIL(mdd_t);
  }

  /* start reached set with initial states */
  reachableStates = scc_mdd_zero(fsm);

  fromLowerBound = mdd_dup(initialStates);
  fromUpperBound = initialStates;
  
  /* Main loop of reachability computation. */
  /* Iterate until fromLowerBound is empty or all states are reached. */
  while ((!mdd_is_tautology(fromLowerBound, 0)) &&
	 (!mdd_is_tautology(reachableStates, 1))) {
    
    #ifdef PRINT
    /* if it is the print step, print out the iteration and the */
    if (printStep && ((depth % printStep) == 0)) {
      int step;
      /* for BFS, preserve earlier format of printing */
      step = Fsm_FsmGetReachableDepth(fsm)+depth;
      PrintStatsPerIteration(approxFlag,
			     nvars,
			     step,
			     stats,
			     mddManager,
			     reachableStates,
			     fromLowerBound,
			     mintermVarArray);
    }
    #endif

    #ifdef REORDER
    if (reorderFlag && !firstReorderFlag &&
	(reachableStateSetSize >= reorderThreshold )){
      firstReorderFlag = TRUE;
      (void) fprintf(vis_stdout, "Dynamic variable ordering forced ");
      (void) fprintf(vis_stdout, "with method sift\n");
      Ntk_NetworkSetDynamicVarOrderingMethod(network,
		     BDD_REORDER_SIFT, BDD_REORDER_VERBOSITY_DEFAULT);
      bdd_reorder(Ntk_NetworkReadMddManager(network));
      Ntk_NetworkSetDynamicVarOrderingMethod(network, currentMethod,
				     BDD_REORDER_VERBOSITY_DEFAULT);
    }
    #endif

    /* careSet for image computation is the set of all states not reached
     * so far.
     */
    toCareSet = mdd_not(reachableStates);

    /*
     * Image of some set between lower and upper, where we care
     * about the image only on unreached states.
     */

    /* compute the image of this iteration */
    image = Img_ImageInfoComputeBwdWithDomainVars(imageInfo,
						    fromLowerBound,
						    fromUpperBound,
						    toCareSet);
    /* free the used bdds */
    mdd_free(toCareSet);
    
    /* New lower bound is the states just reached. */
    mdd_free(fromLowerBound);
    image = scc_MddAndDestroy(image,bound,1,1,1,0);
    fromLowerBound = mdd_and(image, reachableStates, 1, 0);
    depth++;

    mdd_free(image);
    image = NULL;
    
    /* New set of reachable states is old set plus new states in image. */
    reachableStates = (mdd_t *)AddStates(fromLowerBound, reachableStates,
			  FSM_MDD_DONT_FREE, FSM_MDD_FREE);
    fromUpperBound = reachableStates;

    if (mdd_is_tautology(fromLowerBound, 0)) depth--;
      
  } /* end of while loop (while frontier != 0) */


  #ifdef PRINT
  if (mdd_is_tautology(fromLowerBound, 0) || 
      mdd_is_tautology(reachableStates, 1)) {

    if (mdd_is_tautology(reachableStates, 1) && 
       !mdd_is_tautology(fromLowerBound, 0) &&
	printStep && ((depth % printStep) == 0) ) {
      int step;
      /* for BFS, preserve earlier format of printing */
      step = 
      (shellFlag) ? array_n(onionRings) : Fsm_FsmGetReachableDepth(fsm)+depth;
      PrintStatsPerIteration(approxFlag,
			     nvars,
			     step,
			     stats,
			     mddManager,
			     reachableStates,
			     fromLowerBound,
			     mintermVarArray);
    }
  }
  #endif
  
  mdd_free(fromLowerBound);
  /*
  mdd_free(initialStates);
  */

  return reachableStates;
} /* end of scc_computeReachableStates() */
