/**CFile***********************************************************************

  FileName    [sccRecur.c]

  PackageName [scc]

  Synopsis    [Routines to decompose state space into SCCs.]

  Author      [Aiguo Xie]

  Copyright   [Copyright (c) 1999 The University of Southern California.
  All rights reserved.]

******************************************************************************/

#include "fsmInt.h"
#include "bdd.h"
#include "ntm.h"
#include "img.h"
#include "imgInt.h"
#include "mvf.h"
#include "part.h"
#include "sccInt.h"

int sccRecurDepth, maxSccRecurDepth, sccNonSccIterDepth;
double sccTsccNum, sccTsccStateNum;
double sccIsccNum, sccIsccStateNum;
double sccNsccNum, sccNsccStateNum;

extern mdd_t * Mc_ComputeAState_semi_random(  /* by Aiguo */
      mdd_t *set,
      Fsm_Fsm_t *fsm);

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
int scc_ReportSCC(Fsm_Fsm_t *fsm, mdd_t *states, int type, int dir, int record_flag);
mdd_t *scc_NonSCCiteration(Fsm_Fsm_t *fsm,mdd_t *from, mdd_t *bound);
static mdd_t *scc_NonSccBwdIteration(Fsm_Fsm_t *fsm,mdd_t *from, mdd_t *bound);
static mdd_t *scc_NonSccFwdIteration(Fsm_Fsm_t *fsm,mdd_t *from, mdd_t *bound);
mdd_t *scc_FwdEscapingImage(Fsm_Fsm_t *fsm, mdd_t *from);
mdd_t *scc_RestrictedBwdImage(Fsm_Fsm_t *fsm, mdd_t *from, mdd_t *to);
mdd_t *scc_RestrictedFwdImage(Fsm_Fsm_t *fsm, mdd_t *from, mdd_t *to);

mdd_t *scc_TakeAState(Fsm_Fsm_t *fsm, mdd_t *state_set);

static int scc_BackwardThenForwardRecursiveDecompose(
			Fsm_Fsm_t *fsm, mdd_t *seed, mdd_t *stateSpace, int recordSCC);

static int scc_ForwardThenBackwardRecursiveDecompose(
			Fsm_Fsm_t *fsm, mdd_t *seed, mdd_t *stateSpace, int recordSCC);

static int scc_BackwardOnlyRecursiveDecompose(
			Fsm_Fsm_t *fsm, mdd_t *seed, mdd_t *stateSpace, int recordSCC);

static void scc_DecomposeInit(Fsm_Fsm_t *fsm);

void scc_addSCCList(sccMddList **list, mdd_t *states);

/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/
mdd_t *scc_mdd_one(Fsm_Fsm_t *fsm);
mdd_t *scc_mdd_zero(Fsm_Fsm_t *fsm);
int scc_Decompose(Fsm_Fsm_t *fsm, mdd_t *stateSpace, int recordSCC);

/**Function********************************************************************

  Synopsis [Computes SCCs of a state space.]

  Description [Decomposes a state space into SCCs and a set of states
  that do not communicate with any other states. It also determines
  if an SCC is at terminal.
  The idea is to repeatly prune the state space. Each pruning starts
  with a state s that is randomly taken from the remaining state space
  (the entire reachable state space at the beginning). It computes
  B(s), the backward set of s. A recursive routine is then called 
  to compute all SCCs in the set B(s).] 

  SeeAlso [scc_RecursiveBackward()]

******************************************************************************/
int scc_Decompose(Fsm_Fsm_t *fsm, mdd_t *stateSpace, int recordSCC)
{
	mdd_t *S, *s, *Bs, *x;
	double num;

	scc_DecomposeInit(fsm);
	S = mdd_dup(stateSpace);

	if (sccVerbose > 0) {
		num = mdd_count_onset(Part_PartitionReadMddManager(
			fsm->imageInfo->functionData.mddNetwork),
			stateSpace,
			fsm->fsmData.presentStateVars);
		fprintf(vis_stderr, "reachable states: %g\n", num);
	}

	while (!mdd_is_tautology(S,0)) {
		if (sccRecurDepth > 0) {
			fprintf(vis_stderr, 
				" ** error: inconsistent recursion depth.\n");
			return 1;
		}
		s = scc_TakeAState(fsm,S);
		Bs = scc_BackwardReachableStates(fsm,s,S);
		if (sccRecursionType == BACKWARD_ONLY) {
			if (scc_BackwardOnlyRecursiveDecompose(fsm, s, Bs, recordSCC))
				return 1;
		} else  if (sccRecursionType == BACKWARD_FORWARD) {
			if (scc_BackwardThenForwardRecursiveDecompose(fsm,s,Bs, recordSCC))
				return 1;
		} else {
			fprintf(vis_stderr, 
				"** scc_error: invalid recursion type.\n");
			return 1;
		}
		x = scc_MddOrDestroy(s,Bs,1,1,1,1);
		S = scc_MddAndDestroy(S,x,1,0,1,1);
	}
	return 0;
}

void scc_PrintResult(elapseTime, method)
long elapseTime;
int method;
{
	if (method == USC_SCC_METHOD) {
		fprintf(vis_stderr, "\n #nSCC states (%g) in (%g) nSCCs",sccNsccStateNum, sccNsccNum);
		fprintf(vis_stderr, "\n #iSCC states (%g) in (%g) iSCCs",sccIsccStateNum, sccIsccNum);
		fprintf(vis_stderr, "\n #tSCC states (%g) in (%g) tSCCs",sccTsccStateNum, sccTsccNum);
		fprintf(vis_stderr, "\n --------");
		fprintf(vis_stderr, "\n Total states (%g)\n",sccNsccStateNum
							+sccIsccStateNum
							+sccTsccStateNum);
		fprintf(vis_stderr, "\n %-20s%d", "MaxRecurDepth:", 
				maxSccRecurDepth);
	}
	fprintf(vis_stderr, "\n %-20s%g\n", "analysis time:",
				(double)elapseTime/1000.0);
}

/**Function********************************************************************

  Synopsis [Recursively computes SCCs of the backward set of a given state.]

  Description [Recusively decompose the backward set of a state seed into
  SCCs and non-SCC states. Each call to the procedure removes the SCC
  containing the state seed (seed may not be in any SCC at the top call
  of the procedure). Next, all the states that can reach the removed SCC 
  or the seed without passing any states of some undiscovered SCCs are
  iteratively identified and removed. Then, the states of the remaining 
  set that can reach any of the states just removed are identified. 
  This set of states is call the backward frontier set. While the 
  backward frontier set is not empty, the procedure picks a state 
  out of it and compute its backward set, and then calls itself.
  When the call finishes, the backward frontier set gets upated and
  the process repeats until the frontier set is empty.
  ]

  SeeAlso []

******************************************************************************/
static int 
scc_BackwardOnlyRecursiveDecompose(
	Fsm_Fsm_t *fsm, mdd_t *seed, mdd_t *stateSpace, int recordSCC)
{
	mdd_t *Bs, *Fs, *xx, *x, *y, *z, *s, *R, *Bf;

	sccRecurDepth++;
	if (sccRecurDepth > maxSccRecurDepth)
		maxSccRecurDepth = sccRecurDepth;

	s = mdd_dup(seed);
	Bs = mdd_dup(stateSpace);
	Fs = scc_ForwardReachableStates(fsm, s, Bs);

	/* take the intersection */
	x = scc_MddAndDestroy(Fs,Bs,1,1,1,0);

	/* check if the intersection is empty */
	if (!mdd_is_tautology(x,0)) {
		xx = scc_FwdEscapingImage(fsm,x);
		if (mdd_is_tautology(xx,0))
			scc_ReportSCC(fsm,x,TERMINAL, BACKWARD, recordSCC);
		else
			scc_ReportSCC(fsm,x,INTERNAL, BACKWARD, recordSCC);
		mdd_free(xx);
		R = scc_MddAndDestroy(Bs,x,1,0,1,0);
	} else {
		scc_ReportSCC(fsm,s,NONSCC, BACKWARD, 0);
		R = scc_MddAndDestroy(Bs,s,1,0,1,0);
	}
	/* 
	 * iteratively identify all non-SCCs reaching the states
	 * just removed 
	 */
	y = scc_MddOrDestroy(s,x,1,1,1,1);
	z = scc_NonSCCiteration(fsm,y,R);
	if (!mdd_is_tautology(z,0)) {
		scc_ReportSCC(fsm, z, NONSCC, BACKWARD, 0);
		R = scc_MddAndDestroy(R,z,1,0,1,0);
	}
	/* compute the backward frontier set */ 
	xx = scc_MddOrDestroy(y,z,1,1,1,1);
	Bf = scc_RestrictedBwdImage(fsm,xx,R);
	mdd_free(xx);
	
	/* emptiess of Bf <==> emptiness of R */
	while (!mdd_is_tautology(R,0)) { 
		s  = scc_TakeAState(fsm, Bf);
		Bs = scc_BackwardReachableStates(fsm,s,R);
		if (scc_BackwardOnlyRecursiveDecompose(fsm,s,Bs, recordSCC))
			return 1;
		y  = scc_MddOrDestroy(s,Bs,1,1,1,1);
		R  = scc_MddAndDestroy(R, y,1,0,1,0);
		Bf = scc_MddAndDestroy(Bf,y,1,0,1,1);
	}
	/* if Bf is not empty, something is wrong */
	if (!mdd_is_tautology(Bf,0)) {
		fprintf(vis_stderr, 
			"** scc_error: something is wrong with the code.\n");
		return 1;
	}

	mdd_free(Bf);
	mdd_free(R);

	sccRecurDepth--;
	return 0;
}

/**Function********************************************************************

  Synopsis [Recursively computes SCCs of the backward set of a given state.]

  Description [Recusively decompose the backward set of a state seed into
  SCCs and non-SCC states. Each call to the procedure removes the SCC
  containing the state seed (seed may not be in any SCC at the top call
  of the procedure). Next, all the states that can reach the removed SCC 
  or the seed without passing any states of some undiscovered SCCs are
  iteratively identified and removed. Then, the states of the remaining 
  set that can reach any of the states just removed are identified. 
  This set of states is call the backward frontier set. While the 
  backward frontier set is not empty, the procedure picks a state 
  out of it and compute its backward set, and then calls itself.
  When the call finishes, the backward frontier set gets upated and
  the process repeats until the frontier set is empty.]

  SeeAlso [scc_ForwardThenBackwardRecursiveDecompose()]

******************************************************************************/
static int 
scc_BackwardThenForwardRecursiveDecompose(
	Fsm_Fsm_t *fsm, mdd_t *seed, mdd_t *stateSpace, int recordSCC)
{
	mdd_t *Bs, *Fs, *xx, *x, *y, *z, *s, *R, *Bf;

	sccRecurDepth++;
	if (sccRecurDepth > maxSccRecurDepth)
		maxSccRecurDepth = sccRecurDepth;

	s = mdd_dup(seed);
	Bs = mdd_dup(stateSpace);
	Fs = scc_ForwardReachableStates(fsm, s, Bs);

	/* take the intersection */
	x = scc_MddAndDestroy(Fs,Bs,1,1,1,0);

	/* check if the intersection is empty */
	if (!mdd_is_tautology(x,0)) {
		xx = scc_FwdEscapingImage(fsm,x);
		if (mdd_is_tautology(xx,0))
			scc_ReportSCC(fsm,x,TERMINAL, BACKWARD, recordSCC);
		else
			scc_ReportSCC(fsm,x,INTERNAL, BACKWARD, recordSCC);
		mdd_free(xx);
		R = scc_MddAndDestroy(Bs,x,1,0,1,0);
	} else {
		sccNonSccIterDepth = 0;
		scc_ReportSCC(fsm,s,NONSCC, BACKWARD, 0);
		R = scc_MddAndDestroy(Bs,s,1,0,1,0);
	}
	/* 
	 * iteratively identify all non-SCCs reaching the states
	 * just removed 
	 */
	y = scc_MddOrDestroy(s,x,1,1,1,1);

	z = scc_NonSccBwdIteration(fsm,y,R);

	if (!mdd_is_tautology(z,0)) {
		scc_ReportSCC(fsm, z, NONSCC, BACKWARD, 0);
		R = scc_MddAndDestroy(R,z,1,0,1,0);
	}
	/* compute the backward frontier set */ 
	xx = scc_MddOrDestroy(y,z,1,1,1,1);
	Bf = scc_RestrictedBwdImage(fsm,xx,R);
	mdd_free(xx);
	
	/* emptiess of Bf <==> emptiness of R */
	/* bidirection recursion */
	while (!mdd_is_tautology(R,0)) {
		if (mdd_is_tautology(Bf,0)) {
			fprintf(vis_stderr,
			"** scc_error: backward frontier set is empty, but
			the remaining state space is not.\n");
			return 1;
		}
		s = scc_TakeAState(fsm, Bf);
		Fs = scc_ForwardReachableStates(fsm,s,R);	
		if (scc_ForwardThenBackwardRecursiveDecompose(fsm,s,Fs, recordSCC))
			return 1;
		y  = scc_MddOrDestroy(s,Fs,1,1,1,1);
		R  = scc_MddAndDestroy(R,y,1,0,1,0);

		x  = scc_RestrictedBwdImage(fsm,y,R);
		Bf = scc_MddOrDestroy(Bf,x,1,1,1,1);
		Bf = scc_MddAndDestroy(Bf,y,1,0,1,1);
	}
	/* if Bf is not empty, something is wrong */
	if (!mdd_is_tautology(Bf,0)) {
		fprintf(vis_stderr, 
			"** scc_error: something is wrong with the code.\n");
		return 1;
	}
	mdd_free(R);
	mdd_free(Bf);
	sccRecurDepth--;
	return 0;
}

/**Function********************************************************************

  Synopsis [Recursively computes SCCs of the FORWARD set of a given state.]

  Description [Recusively decompose the forward set of a state seed into
  SCCs and non-SCC states. Each call to the procedure removes the SCC
  containing the state seed (seed may not be in any SCC at the top call
  of the procedure). Next, all the states that can be reached only from 
  some already removed states or states that have zero out-degree in the
  remaining set are iteratively identified and removed. Then, the states 
  of the remaining set that can be reached by any of the states just 
  removed are identified. This set of states is call the forward frontier 
  set. If the forward frontier set is not empty, the procedure picks a 
  state out of it and computes its backward set, and calls its dual
  procedure scc_BackwardThenForwardRecursiveDecompose(). When the call 
  returns, the forward frontier set gets upated and the process repeats 
  until the frontier set is empty.]

  SeeAlso [scc_BackwardThenForwardRecursiveDecompose()]

******************************************************************************/
static int 
scc_ForwardThenBackwardRecursiveDecompose(
Fsm_Fsm_t *fsm, mdd_t *seed, mdd_t *stateSpace, int recordSCC)
{
	mdd_t *Bs, *Fs, *xx, *x, *y, *z, *s, *R, *Ff;

	sccRecurDepth++;
	if (sccRecurDepth > maxSccRecurDepth)
		maxSccRecurDepth = sccRecurDepth;

	s = mdd_dup(seed);
	Fs = mdd_dup(stateSpace);
	Bs = scc_BackwardReachableStates(fsm, s, Fs);

	/* take the intersection */
	x = scc_MddAndDestroy(Fs,Bs,1,1,0,1);

	/* check if the intersection is empty */
	if (!mdd_is_tautology(x,0)) { /* the SCC must be internal */ 
		scc_ReportSCC(fsm,x,INTERNAL, FORWARD, recordSCC);
		R = scc_MddAndDestroy(Fs,x,1,0,1,0);
	} else {
		sccNonSccIterDepth = 0;
		scc_ReportSCC(fsm,s,NONSCC, FORWARD, 0);
		R = scc_MddAndDestroy(Fs,s,1,0,1,0);
	}
	/* 
	 * iteratively identify all non-SCCs reaching the states
	 * just removed 
	 */
	y = scc_MddOrDestroy(s,x,1,1,1,1);
	z = scc_NonSccFwdIteration(fsm,y,R);
	if (!mdd_is_tautology(z,0)) {
		scc_ReportSCC(fsm, z, NONSCC, FORWARD, 0);
		R = scc_MddAndDestroy(R,z,1,0,1,0);
	}
	/* compute the forward frontier set */ 
	xx = scc_MddOrDestroy(y,z,1,1,1,1);
	Ff = scc_RestrictedFwdImage(fsm,xx,R);
	mdd_free(xx);
	
	/* emptiess of Ff <==> emptiness of R */
	/* bidirection recursion */
	while (!mdd_is_tautology(R,0)) {
		if (mdd_is_tautology(Ff,0)) {
			fprintf(vis_stderr,
			"** scc_error: forward frontier set is empty, but
			the remaining state space is not.\n");
			return 1;
		}
		s = scc_TakeAState(fsm, Ff);
		Bs = scc_BackwardReachableStates(fsm,s,R);	
		if (scc_BackwardThenForwardRecursiveDecompose(fsm,s,Bs,recordSCC))
			return 1;
		y  = scc_MddOrDestroy(s,Bs,1,1,1,1);
		R  = scc_MddAndDestroy(R,y,1,0,1,0);

		x  = scc_RestrictedFwdImage(fsm,y,R);
		Ff = scc_MddOrDestroy(Ff,x,1,1,1,1);
		Ff = scc_MddAndDestroy(Ff,y,1,0,1,1);
	}
	/* if Ff is not empty, something is wrong */
	if (!mdd_is_tautology(Ff,0)) {
		fprintf(vis_stderr, 
			"** scc_error: something is wrong with the code.\n");
		return 1;
	}
	mdd_free(R);
	mdd_free(Ff);
	sccRecurDepth--;
	return 0;
}

/**Function********************************************************************

  Synopsis [Randomly picks a state from a give set.]

  Description [Current implemention does not guarantees randomness. It
  simplies implements the function Mc_ComputeAState() (vis-1.3).] 

******************************************************************************/
mdd_t *scc_TakeAState(Fsm_Fsm_t *fsm, mdd_t *state_set)
{
	mdd_t *state;
	/*
	state = Mc_ComputeAState(state_set, fsm);
	*/
	state = Mc_ComputeAState_semi_random(state_set, fsm);
	return(state);
}

/**Function********************************************************************

  Synopsis [computes the conjunction of two mdds.]

  Description [computes the conjunction of two mdds.] 

  Side effect [Both mdds are subject to being freed if such specified.]

******************************************************************************/
mdd_t *scc_MddAndDestroy(mdd_t *a, mdd_t *b,
			boolean phase_a, boolean phase_b,
			boolean destory_a, boolean destory_b)
{
	mdd_t *result;
	result = mdd_and(a,b,phase_a,phase_b);
	if (destory_a)
		mdd_free(a);
	if (destory_b)
		mdd_free(b);
	return (result);
}

/**Function********************************************************************

  Synopsis [computes the disjunction of two mdds.]

  Description [computes the disjunction of two mdds.] 

  Side effect [Both mdds are subject to being freed if such specified.]

******************************************************************************/
mdd_t *scc_MddOrDestroy(mdd_t *a, mdd_t *b,
			boolean phase_a, boolean phase_b,
			boolean destory_a, boolean destory_b)
{
	mdd_t *result;
	result = mdd_or(a,b,phase_a,phase_b);
	if (destory_a)
		mdd_free(a);
	if (destory_b)
		mdd_free(b);
	return (result);
}

/**Function********************************************************************

  Synopsis [computes the image from one set to anything beyond that set.] 

  Description [computes the image of set "from" to set "to".] 

  Side effect []

******************************************************************************/
mdd_t *scc_FwdEscapingImage(Fsm_Fsm_t *fsm, mdd_t *from)
{
	mdd_t *image;
	mdd_t *toCareSet;

	toCareSet = mdd_not(from);
	image = Img_ImageInfoComputeFwdWithDomainVars(fsm->imageInfo,
							from,
							from, 
							toCareSet);
	image = scc_MddAndDestroy(image,toCareSet,1,1,1,1);
	return (image);
}

/**Function********************************************************************

  Synopsis [computes the image from one set to another.] 

  Description [computes the image of set "from" to set "to".] 

  Side effect []

******************************************************************************/
mdd_t *scc_RestrictedFwdImage(Fsm_Fsm_t *fsm, mdd_t *from, mdd_t *to)
{
	mdd_t *image;
	image = Img_ImageInfoComputeFwdWithDomainVars(fsm->imageInfo,
							from,
							from, 
							to);
	image = scc_MddAndDestroy(image,to,1,1,1,0);
	return (image);
}

/**Function********************************************************************

  Synopsis [computes the image of one set to another.] 

  Description [computes the backward image of set "from" to set "to".] 

  Side effect []

******************************************************************************/
mdd_t *scc_RestrictedBwdImage(Fsm_Fsm_t *fsm, mdd_t *from, mdd_t *to)
{
	mdd_t *image;
	image = Img_ImageInfoComputeBwdWithDomainVars(fsm->imageInfo,
							from, 
							from, 
							to);
	image = scc_MddAndDestroy(image,to,1,1,1,0);
	return (image);
}


mdd_t *scc_mdd_zero(Fsm_Fsm_t *fsm)
{
	mdd_manager *mddManager;
	mddManager = 
	Part_PartitionReadMddManager(fsm->imageInfo->functionData.mddNetwork);
	return(mdd_zero(mddManager));
}

mdd_t *scc_mdd_one(Fsm_Fsm_t *fsm)
{
	mdd_manager *mddManager;
	mddManager = 
	Part_PartitionReadMddManager(fsm->imageInfo->functionData.mddNetwork);
	return(mdd_one(mddManager));
}

int scc_ReportSCC(Fsm_Fsm_t *fsm, mdd_t *states, int type, int dir, int recordSCC)
{
	mdd_manager *mddManager;
	double num;
	char dirType;
	mddManager = 
	Part_PartitionReadMddManager(fsm->imageInfo->functionData.mddNetwork);

	num = mdd_count_onset(mddManager, states, fsm->fsmData.presentStateVars);
	switch (dir) {
		case BACKWARD:
			dirType = 'B';
			break;
		case FORWARD:
			dirType = 'F';
			break;
		default:
			fprintf(vis_stderr, 
			"** scc: invalid direction to call scc_ReportSCC().\n");
			break;
	}
	switch(type) {
		case INTERNAL:
			if (sccVerbose > 0)
				fprintf(vis_stderr, "iSCC (%12g states). RecurDepth (%7d)%c.\n", num, sccRecurDepth, dirType);
			sccIsccNum += 1;
			sccIsccStateNum += num;
			if (sccVerbose > 0) {
				fprintf(vis_stderr, 
					"# iSCC found (%g)\n", sccIsccNum); 
				fprintf(vis_stderr, 
					"# tSCC found (%g)\n", sccTsccNum); 
				fprintf(vis_stderr, 
					"SCC states (%g)\n", sccIsccStateNum+
							     sccTsccStateNum); 
			}
			if (recordSCC)
				scc_addSCCList(&sccInternalSCCList, states);
			break;
		case TERMINAL:
			if (sccVerbose > 0)
			fprintf(vis_stderr, "tSCC (%12g states). RecurDepth (%7d)%c.\n", num, sccRecurDepth, dirType);
			sccTsccNum += 1;
			sccTsccStateNum += num;
			if (sccVerbose > 0) {
				fprintf(vis_stderr, 
					"# iSCC found (%g)\n", sccIsccNum);
				fprintf(vis_stderr, 
					"# tSCC found (%g)\n", sccTsccNum); 
				fprintf(vis_stderr, 
					"SCC states (%g)\n", sccIsccStateNum+
							     sccTsccStateNum); 
			}
			if (recordSCC)
				scc_addSCCList(&sccExternalSCCList, states);
			break;
		case NONSCC:
			if (sccVerbose > 0)
			fprintf(vis_stderr, "nSCC (%12g states). RecurDepth (%7d). NonSccIterDepth (%7d)%c\n", num, sccRecurDepth, sccNonSccIterDepth, dirType);
			sccNsccNum += 1;
			sccNsccStateNum += num;
			break;
		default:
			fprintf(vis_stderr, "** scc_error: unknown scc type.\n");
			return 1;
	}
	return 0;
}

static void scc_DecomposeInit(Fsm_Fsm_t *fsm) 
{
	fsm->imageInfo = Fsm_FsmReadOrCreateImageInfo(fsm,1,1);

	sccRecurDepth = 0;
	maxSccRecurDepth = sccRecurDepth;
	sccTsccNum = 0;
	sccIsccNum = 0;
	sccNsccNum = 0;
	sccTsccStateNum = 0;
	sccIsccStateNum = 0;
	sccNsccStateNum = 0;
	sccNonSccIterDepth = 0;
}

/**Function********************************************************************

  Synopsis [Computes the set of non-scc states from a given set.] 

  Description [Iteratively computes the set of states that are backward
  reachable from a given set "from" without passing through an scc-state.]

  Side effect []

******************************************************************************/
mdd_t *scc_NonSCCiteration(Fsm_Fsm_t *fsm,mdd_t *from, mdd_t *bound)
{
	mdd_t *x, *y, *frontier, *result, *to;
	Img_ImageInfo_t *imageInfo = NIL(Img_ImageInfo_t); 

	sccNonSccIterDepth = 0;

	frontier = mdd_dup(from);
	to = mdd_dup(bound);

	result = scc_mdd_zero(fsm);

	/*
	if (!(fsm->imageInfo == NIL(Img_ImageInfo_t))) {
		free(fsm->imageInfo);
		fsm->imageInfo = NIL(Img_ImageInfo_t);
	}
	imageInfo = Fsm_FsmReadOrCreateImageInfo(fsm,1,0);
	*/
	imageInfo = fsm->imageInfo;

	while (!mdd_is_tautology(frontier, 0)) {
		sccNonSccIterDepth++;
		x = Img_ImageInfoComputeBwdWithDomainVars(imageInfo,
							frontier, 
							frontier, 
							to);
		x = scc_MddAndDestroy(x,to,1,1,1,0);

		y = Img_ImageInfoComputeBwdWithDomainVars(imageInfo,
							to,
							to,
							x);
		mdd_free(frontier);
		frontier = scc_MddAndDestroy(x,y,1,0,1,1);
		result = scc_MddOrDestroy(result,frontier,1,1,1,0);
		to = scc_MddAndDestroy(to, frontier, 1, 0, 1, 0);
	}
	mdd_free(frontier);
	mdd_free(to);
	return (result);
}

/**Function********************************************************************

  Synopsis [Computes the set of non-scc states from a given set.] 

  Description [Iteratively computes the set of states that are backward
  reachable from a given set "from" without passing through an scc-state.]

  Side effect []

******************************************************************************/
static mdd_t *scc_NonSccBwdIteration(Fsm_Fsm_t *fsm, mdd_t *from, mdd_t *bound)
{
	mdd_t *x, *y, *z, *yy, *zz, *frontier, *result, *to;
	Img_ImageInfo_t *imageInfo = NIL(Img_ImageInfo_t);

	sccNonSccIterDepth = 0;

	if (mdd_is_tautology(from,0)) {
		return(mdd_dup(from));
	}
	frontier = mdd_dup(from);
	to = mdd_dup(bound);

	result = scc_mdd_zero(fsm);

	imageInfo = fsm->imageInfo;

	while (1) {
		x = Img_ImageInfoComputeBwdWithDomainVars(imageInfo,
							frontier, 
							frontier, 
							to);
		x = scc_MddAndDestroy(x,to,1,1,1,0);

		y = Img_ImageInfoComputeBwdWithDomainVars(imageInfo,
							to,
							to,
							x);

		z = Img_ImageInfoComputeFwdWithDomainVars(imageInfo,
							to,
							to,
							x);
		/* for debug */
		zz = scc_MddAndDestroy(x,z,1,0,0,1);
		/* zz = (in-degree-0) states in x */

		yy = scc_MddAndDestroy(x,y,1,0,1,1);
		/* yy = states in x can only reach states in frontier */ 

		mdd_free(frontier);
		frontier = scc_MddOrDestroy(zz,yy,1,1,1,1);
		if (mdd_is_tautology(frontier, 0))
			break;
		sccNonSccIterDepth++;
		result = scc_MddOrDestroy(result,frontier,1,1,1,0);
		to = scc_MddAndDestroy(to, frontier, 1, 0, 1, 0);

		/* for debug
		mdd_free(frontier);
		frontier = scc_MddAndDestroy(x,y,1,0,1,1);
		if (mdd_is_tautology(frontier, 0))
			break;
		sccNonSccIterDepth++;
		result = scc_MddOrDestroy(result,frontier,1,1,1,0);
		to = scc_MddAndDestroy(to, frontier, 1, 0, 1, 0);
		*/
	}
	mdd_free(frontier);
	mdd_free(to);
	return (result);
}

/**Function********************************************************************

  Synopsis [Computes the set of non-scc states from a given set.] 

  Description [Iteratively computes the set of states that are forward
  reachable from a given set "from" with passing through any SCC.]

  Side effect []

******************************************************************************/
mdd_t *scc_NonSccFwdIteration(Fsm_Fsm_t *fsm, mdd_t *from, mdd_t *bound)
{
	mdd_t *x, *y, *z, *yy, *zz, *frontier, *result, *to;
	Img_ImageInfo_t *imageInfo = NIL(Img_ImageInfo_t);

	sccNonSccIterDepth = 0;

	if (mdd_is_tautology(from,0)) {
		return (mdd_dup(from));
	}
	frontier = mdd_dup(from);
	to = mdd_dup(bound);

	result = scc_mdd_zero(fsm);

	imageInfo = fsm->imageInfo;

	while (1) {
		x = Img_ImageInfoComputeFwdWithDomainVars(imageInfo,
							frontier, 
							frontier, 
							to);
		x = scc_MddAndDestroy(x,to,1,1,1,0);

		y = Img_ImageInfoComputeFwdWithDomainVars(imageInfo,
							to,
							to,
							x);

		z = Img_ImageInfoComputeBwdWithDomainVars(imageInfo,
							to,
							to,
							x);
		/* for debug */
		zz = scc_MddAndDestroy(x,z,1,0,0,1);
		/* zz = (out-degree-0) states in x */

		yy = scc_MddAndDestroy(x,y,1,0,1,1);
		/* yy = states in x that can only be reached from frontier */

		mdd_free(frontier);
		frontier = scc_MddAndDestroy(zz,yy,1,1,1,1);
		if (mdd_is_tautology(frontier,0))
			break;
		sccNonSccIterDepth++;
		result = scc_MddOrDestroy(result,frontier,1,1,1,0);
		to = scc_MddAndDestroy(to, frontier, 1, 0, 1, 0);

		/* for debug 
		mdd_free(frontier);
		frontier = scc_MddAndDestroy(x,y,1,0,1,1);
		if (mdd_is_tautology(frontier, 0))
			break;
		sccNonSccIterDepth++;
		result = scc_MddOrDestroy(result,frontier,1,1,1,0);
		to = scc_MddAndDestroy(to, frontier, 1, 0, 1, 0);
		*/
	}
	mdd_free(frontier);
	mdd_free(to);
	return (result);
}

void scc_addSCCList(sccMddList **list, mdd_t *states)
{
	sccMddList *listElement, *curs;

	listElement = (sccMddList *)calloc(1,sizeof(sccMddList));
	listElement->val = mdd_dup(states);
	
	if (*list == NULL)
		*list = listElement;
	else {
		curs = *list;
		while (curs->next) curs = curs->next;
		curs->next = listElement;
	}
	return;
}
