/**CFile***********************************************************************

  FileName    [sccTC.c]

  PackageName [scc]

  Synopsis    [Routine to decompose state space into SCCs based on 
	       transitive closure. It is largely due to the method
	       by University of Corolado at Boulder. Its implementation 
	       is a slight modification of the one provided in the sibling 
	       directory "mark" which picks out only the terminal SCCs.]

  Author      [Aiguo Xie]

  Copyright   [Copyright (c) 1999 The University of Southern California.
  All rights reserved.]

******************************************************************************/

#include "fsmInt.h"
#include "img.h"
#include "imgInt.h"

#include "fsm.h"
#include "bdd.h"
#include "sccInt.h"

static bdd_node * transitiveClosureByIterSquaring(bdd_manager *ddManager, bdd_node *TR, bdd_node **xVars, bdd_node **yVars, bdd_node **zVars, int nVars);
static bdd_node * computeTransitionRelationWithIds(bdd_manager *ddManager, array_t *nextBdds, bdd_node **yVars, int nVars);

/* The following are copied from UCB mark */
static array_t *markGetBddArray(array_t *mvfArray); 
static bdd_node **BddNodeArrayFromIdArray(bdd_manager *ddManager, 
		array_t	*idArray); 
static bdd_node *computeTransitionRelationWithIds(
  bdd_manager	*ddManager,
  array_t	*nextBdds,
  bdd_node	**yVars,
  int		nVars);

static bdd_node *pickOneCube(
  bdd_manager *manager,
  bdd_node *node,
  bdd_node **xVars,
  int nVars);

static bdd_t* mdd2bdd(mdd_manager *mgr, 
	mdd_t *mdd,
	array_t *idArray);

extern bdd_t* mddRetOnvalBdd();

bdd_node **mdd_id_array_to_bdd_vars(
	mdd_manager *mgr,
	array_t *mddIdArray,
	int *nvars);

extern bdd_node *
bdd_get_var(bdd_manager *mgr, int variable_ID);

mdd_t * sccComputeMonoTR(Fsm_Fsm_t* fsm);

long totalSccNum=0;
long totalRecurrentStateNum=0;

int scc_Decompose_TC(Fsm_Fsm_t *fsm, long *initialTime, long *finalTime)
{
    graph_t *partition;
    Ntk_Network_t *network1;
    bdd_manager *ddManager;

    array_t *tranFunArray, *leaveIds; 
    array_t *nextBdds, *nextMvfs;
    array_t *xVarArray, *yVarArray, *piVarArray;

    int i, phase;
    int nVars, nPi;

    bdd_node *zero;
    bdd_node *TR, *new_TR;
    bdd_node **xVars,**yVars, **piVars, **zVars;
    bdd_node *ddTemp, *reachable, *initialStates;
    bdd_node *front, *new_front;

    bdd_node *temp1, *temp2;
    bdd_node *yCube, *xCube, *piCube;
    bdd_node *tc, *cr;
    double numStates;
    double numReachableStates, numRecurrentStates;

    bdd_node *recurrentStates, *seed, *scc;

    mdd_t *reachStates;
    mdd_t *initStates;
    
    if (bdd_get_package_name() != CUDD) {
	(void) fprintf(vis_stderr,"Synthesis package can be used only with CUDD package\n");
	(void) fprintf(vis_stderr,"Please compile with CUDD package\n");
	return 0;
    }

    *initialTime = util_cpu_time();

    ddManager = (bdd_manager *)Fsm_FsmReadMddManager(fsm);

    fsm->imageInfo = Fsm_FsmReadOrCreateImageInfo(fsm,1,1);

    zero = bdd_read_logic_zero(ddManager);

    /*
     * tranFunArray is a list of next state funs.
     */

#ifdef UCB_TR
    network1 = Fsm_FsmReadNetwork(fsm);
    tranFunArray = Fsm_FsmReadNextStateFunctionNames(fsm);

    leaveIds = array_join(Fsm_FsmReadInputVars(fsm),
			  Fsm_FsmReadPresentStateVars(fsm));
    /*
     * Get the BDDs for transition functions.
     Duplicate functions are returned.
     */
    partition = Fsm_FsmReadPartition(fsm);
    nextMvfs = Part_PartitionBuildFunctions(partition,tranFunArray,leaveIds,
					    NIL(mdd_t));
    array_free(leaveIds);
    array_free(tranFunArray);
    nextBdds = markGetBddArray(nextMvfs);

    Mvf_FunctionArrayFree(nextMvfs);

    TR = computeTransitionRelationWithIds(ddManager, nextBdds,
						    yVars, nVars);
    arrayForEachItem(bdd_node *, nextBdds, i, ddTemp) {
 	bdd_recursive_deref(ddManager,ddTemp);
    }
    array_free(nextBdds);

    /* Get the DdNodes for all the variables.*/

    /* ???
    yVars = (bdd_node **)BddNodeArrayFromIdArray(ddManager,
					Fsm_FsmReadNextStateVars(fsm));

    nVars = array_n(Fsm_FsmReadNextStateVars(fsm));

    */
#endif

    /* Compute the transition relation */ 

    if (sccVerbose > 0) {
	fprintf(vis_stderr, "Computing monolithic transition relation...");
    	TR = sccComputeMonoTR(fsm);

    	ddTemp = mdd2bdd(ddManager, TR, 
		array_join(Fsm_FsmReadPresentStateVars(fsm), 
			    Fsm_FsmReadNextStateVars(fsm))); 
    	bdd_recursive_deref(ddManager, TR);
    	TR = ddTemp;

    	ddTemp = (bdd_node *)bdd_get_node(TR,&phase);
    	TR = (phase==TRUE) ?  bdd_not_bdd_node(ddTemp) : ddTemp;
    	bdd_ref(TR);
	fprintf(vis_stderr, "done.\n");
    }

    xVarArray = 
	mdd_id_array_to_bdd_array(ddManager, Fsm_FsmReadPresentStateVars(fsm));
    yVarArray = 
	mdd_id_array_to_bdd_array(ddManager, Fsm_FsmReadNextStateVars(fsm));

    /*
    nVars = array_n(xVarArray);
    xVars = ALLOC(bdd_node *, nVars);
    for (i=0; i<nVars; i++)
	xVars[i] = array_fetch(bdd_t *, xVarArray, i);
    */
    xVars = mdd_id_array_to_bdd_vars(ddManager,
	Fsm_FsmReadPresentStateVars(fsm),&nPi); 
    
    nVars = array_n(yVarArray);
    /*
    yVars = ALLOC(bdd_node *, nVars);
    for (i=0; i<nVars; i++)
	yVars[i] = array_fetch(bdd_t *, yVarArray, i);
    */
    yVars = mdd_id_array_to_bdd_vars(ddManager,
	Fsm_FsmReadNextStateVars(fsm),&nPi); 

    /******/
    initStates = Fsm_FsmComputeInitialStates(fsm);

    ddTemp = mdd2bdd(ddManager, initStates, Fsm_FsmReadPresentStateVars(fsm));
    mdd_free(initStates);
    initStates = ddTemp;

    ddTemp = (bdd_node *)bdd_get_node(initStates,&phase);
    initialStates = (phase==TRUE)? bdd_not_bdd_node(ddTemp) : ddTemp;
    bdd_ref(initialStates);
    mdd_free(initStates);

    if (sccVerbose > 0) {
    	numStates = bdd_count_minterm(ddManager,initialStates,nVars);
    	fprintf(vis_stderr, "initial states: %g.\n", numStates);
    }
    /******/

    /* Compute the reachable states. */
    reachStates = Fsm_FsmComputeReachableStates(fsm,0,sccVerbose,1,0,0,0,1000,
						Fsm_Rch_Default_c,0,0,
						NIL(array_t),FALSE);
    /*
    numStates = 
     mdd_count_onset(ddManager, reachStates, Fsm_FsmReadPresentStateVars(fsm)); 
    */
    /*
    bdd_ref(reachStates); 
    */

    ddTemp = mdd2bdd(ddManager, reachStates, Fsm_FsmReadPresentStateVars(fsm));
    mdd_free(reachStates);
    reachStates = ddTemp;

    ddTemp = (bdd_node *)bdd_get_node(reachStates,&phase);
    reachable = (phase==TRUE) ?  bdd_not_bdd_node(ddTemp) : ddTemp;
    bdd_ref(reachable);
    mdd_free(reachStates);

    numReachableStates = bdd_count_minterm(ddManager,reachable,nVars);
    fprintf(vis_stderr, "reachable states: %g\n", (double)numReachableStates);

    piVars = mdd_id_array_to_bdd_vars(ddManager,Fsm_FsmReadInputVars(fsm),&nPi); 

    /*
    piVarArray =
	mdd_id_array_to_bdd_array(ddManager, Fsm_FsmReadInputVars(fsm));
    nPi = array_n(piVarArray);
    piVars = ALLOC(bdd_node *, nPi);
    for (i=0; i<nPi; i++) {
	bdd_ref(ddTemp = array_fetch(bdd_t *, piVarArray, i));
	piVars[i] = ddTemp;
    }
    */

    bdd_ref(piCube = bdd_bdd_compute_cube(ddManager,piVars,NULL,nPi));

    /* abstract the PI variables for the SCC computation 
    bdd_ref(new_TR = bdd_bdd_exist_abstract(ddManager,TR,piCube));
    bdd_recursive_deref(ddManager,piCube);
    bdd_recursive_deref(ddManager, TR);
    TR = new_TR;
    */

    bdd_ref(xCube = bdd_bdd_compute_cube(ddManager,xVars,NULL,nVars));

#ifdef MONOLITHIC_REACHABILITY
    /* reachability by monolithic TR */
    bdd_ref(front = initialStates); 
    bdd_ref(reachable = front); 
    while (front != zero) {
	bdd_ref(ddTemp = bdd_bdd_and(ddManager, front, TR)); 
	bdd_ref(temp1 = bdd_bdd_exist_abstract(ddManager, ddTemp, xCube));
	bdd_recursive_deref(ddManager, ddTemp);
	bdd_ref(temp2 = 
		bdd_bdd_swap_variables(ddManager, temp1, yVars, xVars, nVars));
	bdd_recursive_deref(ddManager, temp1);
	bdd_ref(new_front = 
		bdd_bdd_and(ddManager, temp2, bdd_not_bdd_node(reachable)));
	bdd_ref(temp1 = bdd_bdd_or(ddManager, reachable, new_front));
	bdd_recursive_deref(ddManager, reachable);
	bdd_recursive_deref(ddManager, temp2);
	reachable = temp1;
	bdd_recursive_deref(ddManager, front);
	front = new_front;

	if (scVerbose > 0) {
        	numStates = bdd_count_minterm(ddManager,new_front,nVars);
        	fprintf(vis_stderr, "|states|: %g\n", numStates);
	}
    }
    bdd_recursive_deref(ddManager, front);
    numStates = bdd_count_minterm(ddManager,reachable,nVars);
    fprintf(vis_stderr, "reachable states: %g\n", numStates);
#endif

    /* Restrict the STG to only the reachale states */
    bdd_ref(ddTemp = bdd_bdd_constrain(ddManager,TR,reachable));
    bdd_recursive_deref(ddManager,TR);
    TR = ddTemp;
    
    /* compute transitive closure of TR */  
    zVars = ALLOC(bdd_node *, nVars);
    for (i = 0; i < nVars; i++)
	bdd_ref(zVars[i] = bdd_bdd_new_var(ddManager));

    if (sccVerbose > 0) {
    	fprintf(vis_stderr, "Computing transitive closure of the transition relation..");
    	tc = transitiveClosureByIterSquaring(ddManager, TR, 
			xVars, yVars, zVars, nVars);
    	bdd_ref(tc);
	fprintf(vis_stderr, "done.\n");
    }


    /* Build the cycle relation CR(x,y) = TC(x,y)*TC(y,x) */
    /* temp1 = tc(y,x) -- retain it to use later in edge relation */
    bdd_ref(temp1 = bdd_bdd_swap_variables(ddManager,tc,xVars,yVars,nVars));
    bdd_ref(cr = bdd_bdd_and(ddManager,tc,temp1));
    bdd_recursive_deref(ddManager, tc);

    bdd_ref(yCube = bdd_bdd_compute_cube(ddManager,yVars,NULL,nVars));
    recurrentStates = bdd_bdd_exist_abstract(ddManager,cr,yCube);

    numRecurrentStates = bdd_count_minterm(ddManager,recurrentStates,nVars);
    if (sccVerbose > 0) {
    	fprintf(vis_stderr, "recurrent states: %g\n", numRecurrentStates);
    }

    /*
    bdd_ref(xCube = bdd_bdd_compute_cube(ddManager,yVars,NULL,nVars));
    */

    totalSccNum=0;
    totalRecurrentStateNum=0;

    *finalTime = util_cpu_time();

    while (recurrentStates != zero) {
	/* pick a state */

	long sccSize;

	/*
	bdd_ref(seed =pickOneCube(ddManager,recurrentStates,xVars,nVars)); 
	*/
	bdd_ref(seed =
	bdd_bdd_pick_one_minterm(ddManager,recurrentStates,xVars,nVars)); 

	numStates = bdd_count_minterm(ddManager, seed, nVars); 
	fprintf(vis_stderr, "seed states: %g\n", numStates);
	/*
	bdd_ref(temp1 = bdd_bdd_constrain(ddManager,cr,seed));
	*/
	bdd_ref(temp1 = bdd_bdd_and(ddManager, cr, seed));
	bdd_ref(temp2 = bdd_bdd_exist_abstract(ddManager, temp1, xCube));
	bdd_recursive_deref(ddManager, temp1);
	bdd_ref(temp1 = 
		bdd_bdd_swap_variables(ddManager, temp2, yVars, xVars, nVars));
	bdd_recursive_deref(ddManager, temp2);

	/* Add the seed to the set, because may be it is not in it */
	bdd_ref(scc = bdd_bdd_or(ddManager,temp1,seed));
	bdd_recursive_deref(ddManager,temp1);

	/* Delete the SCC from the remaining recurrent state set */
	bdd_ref(temp2 = bdd_bdd_and(ddManager,recurrentStates, 
		bdd_not_bdd_node(scc)));
	bdd_recursive_deref(ddManager,recurrentStates); 
	recurrentStates = temp2;

	if (sccVerbose > 0) {
	    sccSize = bdd_count_minterm(ddManager,scc,nVars);
	    fprintf(vis_stderr, "find an SCC with %ld states.\n", sccSize);
	    totalRecurrentStateNum+= sccSize;
	}
	totalSccNum++;
    }
    if (sccVerbose > 0) {
	if (totalRecurrentStateNum != numRecurrentStates) {
		fprintf(vis_stderr, "** scc_error: Numbers of recurrent states do not agree.\n");
		fprintf(vis_stderr, "Please report this error to: aiguo@jungfrau.usc.edu.\n");
		return 1;
	}
    }

    /* clean up */

    bdd_recursive_deref(ddManager, recurrentStates);
    bdd_recursive_deref(ddManager, cr);
    bdd_recursive_deref(ddManager, yCube);
    bdd_recursive_deref(ddManager, xCube);

    /* print result */
    fprintf(vis_stderr, "\n");
    fprintf(vis_stderr, "# reachable states: %g\n", numReachableStates);
    fprintf(vis_stderr, "# recurrent states: %g\n", numRecurrentStates);
    fprintf(vis_stderr, "# SCCs: %d\n", totalSccNum);

    return 0;
}

/**Function********************************************************************

  Synopsis           [Computes the transitive closure by iterative squaring.]

  Description        []

  SideEffects        [None]

  SeeAlso            []

******************************************************************************/
static bdd_node *
transitiveClosureByIterSquaring(
  bdd_manager *ddManager,
  bdd_node *TR,
  bdd_node **xVars,
  bdd_node **yVars,
  bdd_node **zVars,
  int nVars)
{
    bdd_node *prev, *next, *TRxz, *TRzy;
    bdd_node *inter;
    bdd_node *zCube;
  
    bdd_ref(zCube = bdd_bdd_compute_cube(ddManager,zVars,NULL,nVars));

    bdd_ref(prev = TR);
    while(1) {
	TRxz = bdd_bdd_swap_variables(ddManager,prev,yVars,zVars,nVars);
	bdd_ref(TRxz);
	TRzy = bdd_bdd_swap_variables(ddManager,prev,xVars,zVars,nVars);
	bdd_ref(TRzy);
	inter = bdd_bdd_and_abstract(ddManager,TRxz,TRzy,zCube);
	bdd_ref(inter);
	bdd_recursive_deref(ddManager,TRxz);
	bdd_recursive_deref(ddManager,TRzy);
	next = bdd_bdd_or(ddManager,TR,inter);
	bdd_ref(next);
	bdd_recursive_deref(ddManager,inter);
	if(prev == next)
	    break;
	bdd_recursive_deref(ddManager,prev);
	prev = next;
    }
    bdd_recursive_deref(ddManager,zCube);
    bdd_recursive_deref(ddManager,prev);
    bdd_deref(next);
    return next;
}

/**Function********************************************************************

  Synopsis [Returns a BDD array from an Mvf array]

  SideEffects [None]

  SeeAlso     []

******************************************************************************/
static array_t *
markGetBddArray(array_t *mvfArray)
{
    int 	  i,phase;
    array_t *bddArray;
    Mvf_Function_t *mvf;

    bddArray = array_alloc(bdd_node *,0);

    arrayForEachItem(Mvf_Function_t *, mvfArray,i,mvf) {
	mdd_t     *mddTemp;
	bdd_node    *ddNode;

	mddTemp = array_fetch(mdd_t *, mvf, 1);
	ddNode = (bdd_node *) bdd_get_node(mddTemp,&phase);
	bdd_ref(ddNode);

	ddNode = phase ? bdd_not_bdd_node(ddNode) : ddNode;
	array_insert_last(bdd_node *, bddArray, ddNode);
    }
    return bddArray;
}

/**Function********************************************************************

  Synopsis [Returns a BDD array given an integer array of variable indices.]

  SideEffects [None]

  SeeAlso     []

******************************************************************************/
static bdd_node **
BddNodeArrayFromIdArray(
  bdd_manager	*ddManager,
  array_t	*idArray)
{
    bdd_node **xvars;
    int i,id;
    int nvars = array_n(idArray);

    xvars = ALLOC(bdd_node *, nvars);
    if (xvars == NULL)
	return NULL;

    for(i = 0; i < nvars; i++) {
	id = array_fetch(int,idArray,i);
	xvars[i] = bdd_bdd_ith_var(ddManager,id);
	bdd_ref(xvars[i]);
    }
    return xvars;
}

/**Function********************************************************************

  Synopsis [Compute the relation between an array of function and a 
  corresponding array of variables. A BDD is returned which represents
  AND(i=0 -> i<nVars)(yVars[i]==nextBdds). ]

  SideEffects []

  SeeAlso     []

******************************************************************************/
static bdd_node *
computeTransitionRelationWithIds(
  bdd_manager	*ddManager,
  array_t	*nextBdds,
  bdd_node	**yVars,
  int		nVars)
{
    bdd_node	*ddtemp1, *ddtemp2;
    bdd_node	*oldTR, *fn;
    int 		 i;


    oldTR = bdd_read_one(ddManager);

    for(i = 0; i < nVars; i++) {
	ddtemp2  = array_fetch(bdd_node *, nextBdds, i);

	fn = bdd_bdd_xnor(ddManager,ddtemp2,yVars[i]);
	bdd_ref(fn);
	ddtemp1 = bdd_bdd_and(ddManager,oldTR,fn);
	bdd_ref(ddtemp1);
	bdd_recursive_deref(ddManager,fn);
	bdd_recursive_deref(ddManager,oldTR);
	oldTR = ddtemp1;
    }

    return oldTR;
}

/**Function********************************************************************

  Synopsis [Randomly picks a cube from a given function.]

  Description [Randomly picks a cube from a given function.]

  SideEffects [None]

  SeeAlso []

******************************************************************************/
static bdd_node *
pickOneCube(
  bdd_manager *manager,
  bdd_node *node,
  bdd_node **xVars,
  int nVars)
{
    char *string;
    int i, size;
    int *indices;
    bdd_node *old, *new;
  
    size = bdd_num_vars(manager);
    indices = ALLOC(int,nVars);
    string = ALLOC(char,size);
    if(! bdd_bdd_pick_one_cube(manager,node,string)) {
	fprintf(stdout,"mark<->MarkPickOneCube: could not pick a cube\n");
	exit(-1);
    }
    for (i = 0; i < nVars; i++) {
	indices[i] = bdd_node_read_index(xVars[i]);
    }

    bdd_ref(old = bdd_read_one(manager));

    for (i = 0; i < nVars; i++) {
	switch(string[indices[i]]) {
	case 0:
	    bdd_ref(new = bdd_bdd_and(manager,old,bdd_not_bdd_node(xVars[i])));
	    bdd_recursive_deref(manager,old);
	    old = new;
	    break;
	case 1:
	    bdd_ref(new = bdd_bdd_and(manager,old,xVars[i]));
	    bdd_recursive_deref(manager,old);
	    old = new;
	    break;
	}
    }
    FREE(string);
    FREE(indices);
    bdd_deref(old);
    return old;
}

bdd_t *
mdd2bdd(mgr, mdd, idArray)
	mdd_manager *mgr;
	mdd_t *mdd;
	array_t *idArray;
{
	int i, arrSize, mddId;
	bdd_t *onvalBdd, *aOnvalBdd, *res, *tempBdd;

	arrSize = array_n(idArray);
	onvalBdd = bdd_one(mgr);

	for (i = 0; i<arrSize; i++) {
		mddId = array_fetch(int, idArray, i);
		aOnvalBdd = mddRetOnvalBdd(mgr,mddId);

		/*
		tempBdd = bdd_and(onvalBdd, aOnvalBdd,1,1);
		*/
		tempBdd = bdd_and(onvalBdd, aOnvalBdd,1,1);
		bdd_free(onvalBdd);
		bdd_free(aOnvalBdd);
		onvalBdd = tempBdd;
	}
	res = bdd_and(onvalBdd,mdd,1,1);
	bdd_free(onvalBdd);
	return (res);
}

bdd_node **mdd_id_array_to_bdd_vars(mgr,mddIdArray, nvars)
mdd_manager *mgr;
array_t *mddIdArray;
int *nvars;
{
	int i, j;
	array_t	*bddIdArray;
	bdd_node **bdd_vars;

	bddIdArray = mdd_id_array_to_bdd_id_array(mgr,mddIdArray);

	*nvars = array_n(bddIdArray); 
	bdd_vars = (bdd_node **)ALLOC(bdd_node *, *nvars);

	for (i = 0; i < *nvars; i++) {
		j = array_fetch(int, bddIdArray, i);
		bdd_vars[i] = bdd_get_var(mgr, j);
	}
	return (bdd_vars);
}

mdd_t *
sccComputeMonoTR(Fsm_Fsm_t* fsm)
{
	int i;
	mdd_t *monoTR;
	array_t *rootRelations = array_alloc(mdd_t*, 0);

	array_t *rootFunctions; 
	
	graph_t *mddNetwork = Fsm_FsmReadPartition(fsm);
        mdd_manager *mgr = Part_PartitionReadMddManager(mddNetwork);
	array_t *roots = Fsm_FsmReadNextStateFunctionNames(fsm);
        array_t *leaves = array_join(Fsm_FsmReadInputVars(fsm),
			  Fsm_FsmReadPresentStateVars(fsm));
	array_t *rangeVars = Fsm_FsmReadNextStateVars(fsm);

	rootFunctions = Part_PartitionBuildFunctions(mddNetwork,
		roots, leaves, NIL(mdd_t));

	array_free(leaves);

	/* 
	 * Take the product of the transition relation of all the
	 * root nodes and smooth out the quantify variables (PIs).
	 */

	/*
	 * Iterate over the function of all the root nodes.
	 */
	 for (i = 0; i<array_n(rootFunctions); i++) {
		Mvf_Function_t *rootFunction = array_fetch(Mvf_Function_t*,
			rootFunctions, i);
		int rangeVarMddId = array_fetch(int, rangeVars, i);
		mdd_t *rootRelation = 
			Mvf_FunctionBuildRelationWithVariable(rootFunction,
			rangeVarMddId);
		array_insert_last(mdd_t*, rootRelations, rootRelation);
	}

	Mvf_FunctionArrayFree(rootFunctions);
	monoTR = Img_MultiwayLinearAndSmooth(mgr, rootRelations,
				Fsm_FsmReadInputVars(fsm),
				Fsm_FsmReadPresentStateVars(fsm));
	mdd_array_free(rootRelations);
	return (monoTR);
}

